using Microsoft.EntityFrameworkCore;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using Elwig.Helpers;
using Elwig.Models;
using System.Threading.Tasks;
using Microsoft.EntityFrameworkCore.ChangeTracking;
using System.Collections.ObjectModel;
using Elwig.Documents;

namespace Elwig.Windows {
    public partial class MemberAdminWindow : AdministrationWindow {

        private List<string> TextFilter = new();
        private readonly RoutedCommand CtrlF = new();
        private readonly (ComboBox, TextBox, TextBox)[] PhoneNrInputs;

        private static ObservableCollection<KeyValuePair<string, string>> PhoneNrTypes { get; set; } = new() {
            new("landline", "Tel.-Nr. (Festnetz)"),
            new("mobile", "Tel.-Nr. (mobil)"),
            new("fax", "Fax-Nr."),
        };

        public MemberAdminWindow() {
            InitializeComponent();
            CtrlF.InputGestures.Add(new KeyGesture(Key.F, ModifierKeys.Control));
            CommandBindings.Add(new CommandBinding(CtrlF, FocusSearchInput));
            ExemptInputs = new Control[] {
                SearchInput, ActiveMemberInput, MemberList,
            };
            RequiredInputs = new Control[] {
                MgNrInput, GivenNameInput, FamilyNameInput,
                AddressInput, PlzInput, OrtInput, BillingOrtInput,
                BusinessSharesInput, BranchInput, DefaultKgInput
            };
            PhoneNrInputs = new (ComboBox, TextBox, TextBox)[] {
                (PhoneNr1TypeInput, PhoneNr1Input, PhoneNr1CommentInput),
                (PhoneNr2TypeInput, PhoneNr2Input, PhoneNr2CommentInput),
                (PhoneNr3TypeInput, PhoneNr3Input, PhoneNr3CommentInput),
                (PhoneNr4TypeInput, PhoneNr4Input, PhoneNr4CommentInput),
                (PhoneNr5TypeInput, PhoneNr5Input, PhoneNr5CommentInput),
                (PhoneNr6TypeInput, PhoneNr6Input, PhoneNr6CommentInput),
                (PhoneNr7TypeInput, PhoneNr7Input, PhoneNr7CommentInput),
                (PhoneNr8TypeInput, PhoneNr8Input, PhoneNr8CommentInput),
                (PhoneNr9TypeInput, PhoneNr9Input, PhoneNr9CommentInput),
            };
            foreach (var input in PhoneNrInputs) input.Item1.ItemsSource = PhoneNrTypes;

            InitializeDelayTimer(SearchInput, SearchInput_TextChanged);
            SearchInput.TextChanged -= SearchInput_TextChanged;
        }

        private void Window_Loaded(object sender, RoutedEventArgs evt) {
            Menu_Print_Letterheads_MgNr.IsEnabled = App.IsPrintingReady;
            Menu_Print_Letterheads_Name.IsEnabled = App.IsPrintingReady;

            ActiveMemberInput.IsChecked = true;
            UpdatePhoneNrInputVisibility();
            LockInputs();
        }

        private async Task RefreshMemberList() {
            await Context.Members.LoadAsync();
            await RefreshMemberListQuery();
        }

        private async Task RefreshMemberListQuery(bool updateSort = false) {
            IQueryable<Member> memberQuery = Context.Members;
            if (ActiveMemberInput.IsChecked == true)
                memberQuery = memberQuery.Where(m => m.IsActive);

            List<Member> members = await memberQuery.ToListAsync();

            if (TextFilter.Count > 0) {
                var dict = members.AsParallel()
                    .ToDictionary(m => m, m => m.SearchScore(TextFilter))
                    .OrderByDescending(a => a.Value)
                    .ThenBy(a => a.Key.FamilyName)
                    .ThenBy(a => a.Key.GivenName);
                var threshold = dict.Select(a => a.Value).Max() * 3 / 4;
                members = dict
                    .Where(a => a.Value > threshold)
                    .Select(a => a.Key)
                    .ToList();
            } else {
                members = members
                    .OrderBy(m => m.FamilyName)
                    .ThenBy(m => m.GivenName)
                    .ToList();
            }

            ControlUtils.RenewItemsSource(MemberList, members, i => (i as Member)?.MgNr,
                MemberList_SelectionChanged, TextFilter.Count > 0 ? ControlUtils.RenewSourceDefault.IfOnly : ControlUtils.RenewSourceDefault.None, !updateSort);
        }

        private void RefreshInputs(bool validate = false) {
            ClearInputStates();
            if (MemberList.SelectedItem is Member m) {
                EditMemberButton.IsEnabled = true;
                DeleteMemberButton.IsEnabled = true;
                AreaCommitmentButton.IsEnabled = true;
                DeliveryButton.IsEnabled = true;
                FillInputs(m);
            } else {
                EditMemberButton.IsEnabled = false;
                DeleteMemberButton.IsEnabled = false;
                AreaCommitmentButton.IsEnabled = false;
                DeliveryButton.IsEnabled = false;
                ClearOriginalValues();
                ClearDefaultValues();
                ClearInputs(validate);
                ClearInputStates();
            }
            GC.Collect();
        }

        private async void InitInputs() {
            ClearOriginalValues();
            ClearDefaultValues();

            MgNrInput.Text = (await Context.NextMgNr()).ToString();
            EntryDateInput.Text = DateTime.Now.ToString("dd.MM.yyyy");
            if (Context.Branches.Count() == 1)
                BranchInput.SelectedItem = Context.Branches.First();
            ActiveInput.IsChecked = true;
            ContactPostalInput.IsChecked = true;

            SetDefaultValue(MgNrInput);
            SetDefaultValue(EntryDateInput);
            SetDefaultValue(ActiveInput);
            ValidateRequiredInputs();
        }

        protected override async Task OnRenewContext() {
            await base.OnRenewContext();
            ControlUtils.RenewItemsSource(BranchInput, await Context.Branches.OrderBy(b => b.Name).ToListAsync(), i => (i as Branch)?.ZwstId);
            ControlUtils.RenewItemsSource(DefaultKgInput, await Context.WbKgs.Select(k => k.AtKg).OrderBy(k => k.Name).ToListAsync(), i => (i as AT_Kg)?.KgNr);
            await RefreshMemberList();
            StatusMembers.Text = $"Mitglieder: {await Context.Members.CountAsync(m => m.IsActive):N0} ({await Context.Members.CountAsync():N0})";
            StatusBusinessShares.Text = $"Geschäftsanteile: {await Context.Members.Where(m => m.IsActive).SumAsync(m => m.BusinessShares):N0} ({await Context.Members.SumAsync(m => m.BusinessShares):N0})";
        }

        private void SetPhoneNrInput(int nr, string? type, string? number, string? comment) {
            var inputs = PhoneNrInputs[nr];
            inputs.Item1.SelectedItem = (type == null) ? null : inputs.Item1.ItemsSource.Cast<KeyValuePair<string, string>>().FirstOrDefault(p => p.Key == type);
            inputs.Item2.Text = number;
            inputs.Item3.Text = comment;
        }

        private (string, string, string?)? GetPhoneNrInput(int nr) {
            var inputs = PhoneNrInputs[nr];
            var number = inputs.Item2.Text;
            if (string.IsNullOrEmpty(number))
                return null;
            var type = (inputs.Item1.SelectedItem as KeyValuePair<string, string>?)?.Key ?? (number.StartsWith("+43 ") && number[4] == '6' ? "mobile" : "landline");
            var comment = inputs.Item3.Text;
            return (type, number, comment == "" ? null : comment);
        }

        private void SetPhoneNrInputVisible(int nr, bool visible) {
            var inputs = PhoneNrInputs[nr];
            var vis = visible ? Visibility.Visible : Visibility.Hidden;
            inputs.Item1.Visibility = vis;
            inputs.Item2.Visibility = vis;
            inputs.Item3.Visibility = vis;
        }

        private void MemberList_SelectionChanged(object sender, RoutedEventArgs evt) {
            RefreshInputs();
        }

        private async void ActiveMemberInput_Changed(object sender, RoutedEventArgs evt) {
            await RefreshMemberListQuery();
        }

        private void NewMemberButton_Click(object sender, RoutedEventArgs evt) {
            IsCreating = true;
            MemberList.IsEnabled = false;
            MemberList.SelectedItem = null;
            HideNewEditDeleteButtons();
            ShowSaveResetCancelButtons();
            UnlockInputs();
            UpdatePhoneNrInputVisibility(true);
            InitInputs();
            LockSearchInputs();
        }

        private void EditMemberButton_Click(object sender, RoutedEventArgs evt) {
            if (MemberList.SelectedItem == null)
                return;

            IsEditing = true;
            MemberList.IsEnabled = false;

            HideNewEditDeleteButtons();
            ShowSaveResetCancelButtons();
            UnlockInputs();
            UpdatePhoneNrInputVisibility(true);
            LockSearchInputs();
        }

        private async void DeleteMemberButton_Click(object sender, RoutedEventArgs evt) {
            if (MemberList.SelectedItem is not Member m)
                return;

            var r = MessageBox.Show(
                $"Soll das Mitglied \"{m.AdministrativeName}\" (MgNr. {m.MgNr}) wirklich unwiderruflich gelöscht werden?",
                "Mitglied löschen", MessageBoxButton.YesNo, MessageBoxImage.Warning, MessageBoxResult.No);
            if (r == MessageBoxResult.Yes) {
                Context.Remove(m);
                await Context.SaveChangesAsync();
                await RefreshMemberList();
            }
        }

        private async void SaveButton_Click(object sender, RoutedEventArgs evt) {
            Member m = await UpdateMember(MemberList.SelectedItem as Member);
            IsEditing = false;
            IsCreating = false;
            MemberList.IsEnabled = true;
            HideSaveResetCancelButtons();
            ShowNewEditDeleteButtons();
            LockInputs();
            UpdatePhoneNrInputVisibility();
            UnlockSearchInputs();
            FinishInputFilling();
            await RefreshMemberList();
            RefreshInputs();
            SearchInput.Text = "";
            MemberList.SelectedItem = m;
        }

        private void ResetButton_Click(object sender, RoutedEventArgs evt) {
            if (IsEditing) {
                RefreshInputs();
            } else if (IsCreating) {
                ClearInputs();
                InitInputs();
            }
            UpdateButtons();
        }

        private void CancelButton_Click(object sender, RoutedEventArgs evt) {
            IsEditing = false;
            IsCreating = false;
            MemberList.IsEnabled = true;
            HideSaveResetCancelButtons();
            ShowNewEditDeleteButtons();
            RefreshInputs();
            LockInputs();
            UpdatePhoneNrInputVisibility();
            UnlockSearchInputs();
        }

        private void AreaCommitmentButton_Click(object sender, RoutedEventArgs evt) {
            var w = new AreaComAdminWindow(((Member)MemberList.SelectedItem).MgNr);
            w.Show();
        }

        private void DeliveryButton_Click(object sender, RoutedEventArgs evt) {
            var w = new DeliveryAdminWindow(((Member)MemberList.SelectedItem).MgNr);
            w.Show();
        }

        private async void SearchInput_TextChanged(object sender, RoutedEventArgs evt) {
            TextFilter = SearchInput.Text.ToLower().Split(" ").ToList().FindAll(e => e.Length > 0);
            await RefreshMemberListQuery(true);
        }

        private void Menu_Member_SendEmail_Click(object sender, RoutedEventArgs evt) {
            if (MemberList.SelectedItem is not Member m)
                return;
            Utils.MailTo(m.EmailAddresses.Select(a => a.Address));
        }

        private async void Menu_Print_Letterhead_Click(object sender, RoutedEventArgs evt) {
            if (MemberList.SelectedItem is not Member m)
                return;
            Mouse.OverrideCursor = Cursors.AppStarting;
            using var doc = new Letterhead(m);
            await doc.Generate();
            Mouse.OverrideCursor = null;
            if (App.Config.Debug) {
                doc.Show();
            } else {
                await doc.Print();
            }
        }

        private async void Menu_Print_Letterheads_MgNr_Click(object sender, RoutedEventArgs evt) {
            var n = await Context.Members.CountAsync(m => m.IsActive);
            var res = MessageBox.Show(
                $"Sollen wirklich {n} Seiten gedruckt werden?", "Ausdruck Bestätigen",
                MessageBoxButton.YesNo, MessageBoxImage.Warning, MessageBoxResult.No);
            if (res != MessageBoxResult.Yes)
                return;
            Mouse.OverrideCursor = Cursors.AppStarting;
            using var doc = await Document.Merge(Context.Members
                .Where(m => m.IsActive && m.ContactViaPost)
                .OrderBy(m => m.MgNr)
                .Select(m => new Letterhead(m)));
            await doc.Generate();
            Mouse.OverrideCursor = null;
            if (App.Config.Debug) {
                doc.Show();
            } else {
                await doc.Print();
            }
        }

        private async void Menu_Print_Letterheads_Name_Click(object sender, RoutedEventArgs evt) {
            var n = await Context.Members.CountAsync(m => m.IsActive);
            var res = MessageBox.Show(
                $"Sollen wirklich {n} Seiten gedruckt werden?", "Ausdruck Bestätigen",
                MessageBoxButton.YesNo, MessageBoxImage.Warning, MessageBoxResult.No);
            if (res != MessageBoxResult.Yes)
                return;
            Mouse.OverrideCursor = Cursors.AppStarting;
            using var doc = await Document.Merge(Context.Members
                .Where(m => m.IsActive && m.ContactViaPost)
                .OrderBy(m => m.FamilyName)
                .ThenBy(m => m.GivenName)
                .Select(m => new Letterhead(m)));
            await doc.Generate();
            Mouse.OverrideCursor = null;
            if (App.Config.Debug) {
                doc.Show();
            } else {
                await doc.Print();
            }
        }

        private void FocusSearchInput(object sender, RoutedEventArgs evt) {
            if (!IsEditing && !IsCreating) {
                SearchInput.Focus();
                SearchInput.SelectAll();
            }
        }

        private void ShowSaveResetCancelButtons() {
            SaveButton.IsEnabled = false;
            ResetButton.IsEnabled = false;
            CancelButton.IsEnabled = true;
            SaveButton.Visibility = Visibility.Visible;
            ResetButton.Visibility = Visibility.Visible;
            CancelButton.Visibility = Visibility.Visible;
        }

        private void HideSaveResetCancelButtons() {
            SaveButton.IsEnabled = false;
            ResetButton.IsEnabled = false;
            CancelButton.IsEnabled = false;
            SaveButton.Visibility = Visibility.Hidden;
            ResetButton.Visibility = Visibility.Hidden;
            CancelButton.Visibility = Visibility.Hidden;
        }

        private void ShowNewEditDeleteButtons() {
            NewMemberButton.IsEnabled = true;
            EditMemberButton.IsEnabled = MemberList.SelectedItem != null;
            DeleteMemberButton.IsEnabled = MemberList.SelectedItem != null;
            NewMemberButton.Visibility = Visibility.Visible;
            EditMemberButton.Visibility = Visibility.Visible;
            DeleteMemberButton.Visibility = Visibility.Visible;
        }

        private void HideNewEditDeleteButtons() {
            NewMemberButton.IsEnabled = false;
            EditMemberButton.IsEnabled = false;
            DeleteMemberButton.IsEnabled = false;
            NewMemberButton.Visibility = Visibility.Hidden;
            EditMemberButton.Visibility = Visibility.Hidden;
            DeleteMemberButton.Visibility = Visibility.Hidden;
        }

        private void LockSearchInputs() {
            SearchInput.IsEnabled = false;
            ActiveMemberInput.IsEnabled = false;
        }

        private void UnlockSearchInputs() {
            SearchInput.IsEnabled = true;
            ActiveMemberInput.IsEnabled = true;
        }

        private void UpdatePhoneNrInputVisibility(bool extra = false) {
            bool lastVisible = true;
            var m = (Member)MemberList.SelectedItem;
            for (int i = 0; i < PhoneNrInputs.Length; i++) {
                var input = PhoneNrInputs[i];
                var vis = !string.IsNullOrEmpty(input.Item2.Text) || (m?.TelephoneNumbers.Any(p => p.Nr - 1 == i) ?? false);
                SetPhoneNrInputVisible(i, vis || (extra && lastVisible));
                lastVisible = vis;
            }
        }

        private async Task<Member> UpdateMember(Member? m) {
            bool memberNew = (m == null);
            m ??= Context.CreateProxy<Member>();

            int newMgNr = int.Parse(MgNrInput.Text);
            m.PredecessorMgNr = (PredecessorMgNrInput.Text == "") ? null : int.Parse(PredecessorMgNrInput.Text);
            m.Prefix = (PrefixInput.Text == "") ? null : PrefixInput.Text;
            m.GivenName = GivenNameInput.Text;
            m.FamilyName = FamilyNameInput.Text;
            m.Suffix = (SuffixInput.Text == "") ? null : SuffixInput.Text;
            m.Birthday = (BirthdayInput.Text == "") ? null : string.Join("-", BirthdayInput.Text.Split(".").Reverse());
            m.IsDeceased = DeceasedInput.IsChecked ?? false;
            m.CountryNum = 40;  // Austria AT AUT
            m.PostalDestId = ((AT_PlzDest)OrtInput.SelectedItem).Id;
            m.Address = AddressInput.Text;

            m.Iban = (IbanInput.Text == "") ? null : IbanInput.Text.Replace(" ", "");
            m.Bic = (BicInput.Text == "") ? null : BicInput.Text;

            m.UstIdNr = (UstIdNrInput.Text == "") ? null : UstIdNrInput.Text;
            m.LfbisNr = (LfbisNrInput.Text == "") ? null : LfbisNrInput.Text;
            m.IsBuchführend = BuchführendInput.IsChecked ?? false;

            m.EntryDateString = (EntryDateInput.Text == "") ? null : string.Join("-", EntryDateInput.Text.Split(".").Reverse());
            m.ExitDateString = (ExitDateInput.Text == "") ? null : string.Join("-", ExitDateInput.Text.Split(".").Reverse());
            m.BusinessShares = (BusinessSharesInput.Text == "") ? 0 : int.Parse(BusinessSharesInput.Text);
            m.AccountingNr = (AccountingNrInput.Text == "") ? null : AccountingNrInput.Text;
            m.IsActive = ActiveInput.IsChecked ?? false;
            m.IsVollLieferant = VollLieferantInput.IsChecked ?? false;
            m.IsFunktionär = FunkionärInput.IsChecked ?? false;
            m.ZwstId = ((Branch)BranchInput.SelectedItem).ZwstId;
            m.DefaultKgNr = ((AT_Kg)DefaultKgInput.SelectedItem).KgNr;
            m.Comment = (CommentInput.Text == "") ? null : CommentInput.Text;
            m.ContactViaPost = ContactPostalInput.IsChecked ?? false;
            m.ContactViaEmail = ContactEmailInput.IsChecked ?? false;

            EntityEntry<Member>? tr = null;
            try {
                if (IsEditing) {
                    tr = Context.Update(m);
                } else if (IsCreating) {
                    m.MgNr = newMgNr;
                    tr = (await Context.AddAsync(m));
                } else {
                    throw new Exception();
                }

                if (BillingOrtInput.SelectedItem == null) {
                    if (m.BillingAddress != null) {
                        Context.Remove(m.BillingAddress);
                    }
                } else {
                    BillingAddr b = m.BillingAddress ?? Context.CreateProxy<BillingAddr>();
                    b.Name = BillingNameInput.Text;
                    b.Address = BillingAddressInput.Text;
                    var p = (AT_PlzDest)BillingOrtInput.SelectedItem;
                    b.CountryNum = p.CountryNum;
                    b.PostalDestId = p.Id;
                    if (m.BillingAddress == null) {
                        b.MgNr = newMgNr;
                        await Context.AddAsync(b);
                    } else {
                        Context.Update(b);
                    }
                }

                for (int i = 0, j = 0; i < PhoneNrInputs.Length; i++) {
                    var input = GetPhoneNrInput(i);
                    var phoneNr = m.TelephoneNumbers.FirstOrDefault(p => p.Nr - 1 == i);
                    if (input == null) {
                        if (phoneNr != null) {
                            Context.Remove(phoneNr);
                        }
                    } else {
                        var pInput = input.Value;
                        MemberTelNr p = phoneNr ?? Context.CreateProxy<MemberTelNr>();
                        p.Nr = ++j;
                        p.Type = pInput.Item1;
                        p.Number = pInput.Item2;
                        p.Comment = pInput.Item3;
                        if (phoneNr == null) {
                            p.MgNr = newMgNr;
                            await Context.AddAsync(p);
                        } else {
                            Context.Update(p);
                        }
                    }
                }

                for (int i = 0; i < 2; i++) {
                    var input = i == 0 ? EmailAddress1Input : EmailAddress2Input;
                    var emailAddr = m.EmailAddresses.FirstOrDefault(a => a.Nr - 1 == i);
                    if (input.Text == "") {
                        if (emailAddr != null) {
                            Context.Remove(emailAddr);
                        }
                    } else {
                        MemberEmailAddr a = emailAddr ?? Context.CreateProxy<MemberEmailAddr>();
                        a.Nr = i + 1;
                        a.Address = input.Text;
                        a.Comment = null;
                        if (emailAddr == null) {
                            a.MgNr = newMgNr;
                            await Context.AddAsync(a);
                        } else {
                            Context.Update(a);
                        }
                    }
                }

                await Context.SaveChangesAsync();

                if (newMgNr != m.MgNr) {
                    await Context.Database.ExecuteSqlAsync($"UPDATE member SET mgnr = {newMgNr} WHERE mgnr = {m.MgNr}");
                    tr.State = EntityState.Detached;
                    await Context.SaveChangesAsync();
                    await tr.ReloadAsync();
                    m = await Context.Members.FindAsync(newMgNr);
                }
            } catch (Exception exc) {
                if (tr != null) {
                    tr.State = EntityState.Detached;
                    await tr.ReloadAsync();
                }
                var str = "Der Eintrag konnte nicht in der Datenbank aktualisiert werden!\n\n" + exc.Message;
                if (exc.InnerException != null) str += "\n\n" + exc.InnerException.Message;
                MessageBox.Show(str, "Mitglied aktualisieren", MessageBoxButton.OK, MessageBoxImage.Error);
            }

            return m;
        }

        private void FillInputs(Member m) {
            ClearOriginalValues();
            ClearDefaultValues();

            MgNrInput.Text = m.MgNr.ToString();
            PredecessorMgNrInput.Text = m.PredecessorMgNr.ToString();
            PrefixInput.Text = m.Prefix;
            GivenNameInput.Text = m.GivenName;
            FamilyNameInput.Text = m.FamilyName;
            SuffixInput.Text = m.Suffix;
            BirthdayInput.Text = (m.Birthday != null) ? string.Join(".", m.Birthday.Split("-").Reverse()) : null;
            if (m.Birthday?.Length == 10) {
                Age.Text = Utils.GetAge(DateOnly.ParseExact(m.Birthday, "yyyy-MM-dd")).ToString();
            } else if (m.Birthday != null) {
                Age.Text = "ca. " + (DateTime.Now.Year - int.Parse(m.Birthday[^4..])).ToString();
            } else {
                Age.Text = "-";
            }
            DeceasedInput.IsChecked = m.IsDeceased;
            AddressInput.Text = m.Address;
            AT_PlzDest? p = m.PostalDest.AtPlz;
            if (p != null) {
                PlzInput.Text = p.Plz.ToString();
                OrtInput.SelectedItem = p;
            } else {
                PlzInput.Text = null;
                OrtInput.SelectedItem = null;
            }

            var emailAddrs = m.EmailAddresses.OrderBy(a => a.Nr).ToList();
            EmailAddress1Input.Text = emailAddrs.Count > 0 ? emailAddrs[0].Address : "";
            EmailAddress2Input.Text = emailAddrs.Count > 1 ? emailAddrs[1].Address : "";

            var phoneNrs = m.TelephoneNumbers.OrderBy(p => p.Nr).ToList();
            for (int i = 0; i < PhoneNrInputs.Length; i++) {
                if (i < phoneNrs.Count) {
                    var phoneNr = phoneNrs[i];
                    SetPhoneNrInput(i, phoneNr.Type, phoneNr.Number, phoneNr.Comment);
                } else {
                    SetPhoneNrInput(i, null, null, null);
                }
            }
            UpdatePhoneNrInputVisibility(IsEditing || IsCreating);

            IbanInput.Text = m.Iban;
            BicInput.Text = m.Bic;

            UstIdNrInput.Text = m.UstIdNr;
            LfbisNrInput.Text = m.LfbisNr;
            BuchführendInput.IsChecked = m.IsBuchführend;

            var billingAddr = m.BillingAddress;
            if (billingAddr != null) {
                BillingNameInput.Text = billingAddr.Name;
                BillingAddressInput.Text = billingAddr.Address;
                AT_PlzDest? b = billingAddr.PostalDest.AtPlz;
                if (b != null) {
                    BillingPlzInput.Text = b.Plz.ToString();
                    BillingOrtInput.SelectedItem = b;
                }
            } else {
                BillingNameInput.Text = "";
                BillingAddressInput.Text = "";
                BillingPlzInput.Text = "";
                BillingOrtInput.SelectedItem = null;
            }

            EntryDateInput.Text = (m.EntryDateString != null) ? string.Join(".", m.EntryDateString.Split("-").Reverse()) : null;
            ExitDateInput.Text = (m.ExitDateString != null) ? string.Join(".", m.ExitDateString.Split("-").Reverse()) : null;
            BusinessSharesInput.Text = m.BusinessShares.ToString();
            AccountingNrInput.Text = m.AccountingNr;
            BranchInput.SelectedItem = m.Branch;
            DefaultKgInput.SelectedItem = m.DefaultKg;
            CommentInput.Text = m.Comment;
            ActiveInput.IsChecked = m.IsActive;
            VollLieferantInput.IsChecked = m.IsVollLieferant;
            FunkionärInput.IsChecked = m.IsFunktionär;
            ContactPostalInput.IsChecked = m.ContactViaPost;
            ContactEmailInput.IsChecked = m.ContactViaEmail;

            var d1 = Context.Deliveries.Where(d => d.Year == Utils.CurrentLastSeason && d.MgNr == m.MgNr);
            var d2 = Context.Deliveries.Where(d => d.Year == Utils.CurrentLastSeason - 1 && d.MgNr == m.MgNr);
            StatusDeliveriesLastSeason.Text = $"Lieferungen ({Utils.CurrentLastSeason - 1}): {d2.Count():N0} ({d2.Sum(d => d.Parts.Count):N0}), {d2.SelectMany(d => d.Parts).Sum(p => p.Weight):N0} kg";
            StatusDeliveriesThisSeason.Text = $"Lieferungen ({Utils.CurrentLastSeason}): {d1.Count():N0} ({d1.Sum(d => d.Parts.Count):N0}), {d1.SelectMany(d => d.Parts).Sum(p => p.Weight):N0} kg";
            StatusAreaCommitment.Text = $"Gebundene Fläche: {m.ActiveAreaCommitments.Select(c => c.Area).Sum():N0} m²";

            Menu_Member_SendEmail.IsEnabled = m.EmailAddresses.Count > 0;
            Menu_Print_Letterhead.IsEnabled = true;

            FinishInputFilling();
        }

        new protected void ClearInputs(bool validate = false) {
            Menu_Member_SendEmail.IsEnabled = false;
            Menu_Print_Letterhead.IsEnabled = false;
            StatusDeliveriesLastSeason.Text = $"Lieferungen ({Utils.CurrentLastSeason - 1}): -";
            StatusDeliveriesThisSeason.Text = $"Lieferungen ({Utils.CurrentLastSeason}): -";
            StatusAreaCommitment.Text = "Gebundene Fläche: -";
            Age.Text = "-";
            base.ClearInputs(validate);
        }

        override protected void UpdateButtons() {
            if (!IsEditing && !IsCreating) return;
            bool ch = HasChanged, v = IsValid;
            ResetButton.IsEnabled = ch;
            SaveButton.IsEnabled = v && ch;
        }

        protected void InputTextChanged(TextBox input, Func<TextBox, bool, AppDbContext, Member?, ValidationResult> checker) {
            InputTextChanged(input, checker(input, SenderIsRequired(input), Context, (Member)MemberList.SelectedItem));
        }

        protected void InputLostFocus(TextBox input, Func<TextBox, bool, AppDbContext, Member?, ValidationResult> checker, string? msg = null) {
            InputLostFocus(input, checker(input, SenderIsRequired(input), Context, (Member)MemberList.SelectedItem), msg);
        }

        private void MgNrInput_TextChanged(object sender, RoutedEventArgs evt) {
            InputTextChanged((TextBox)sender, Validator.CheckNewMgNr);
        }

        private void MgNrInput_LostFocus(object sender, RoutedEventArgs evt) {
            InputLostFocus((TextBox)sender, Validator.CheckNewMgNr);
        }

        private void PredecessorMgNrInput_TextChanged(object sender, RoutedEventArgs evt) {
            InputTextChanged((TextBox)sender, Validator.CheckPredecessorMgNr);
        }

        private void PredecessorMgNrInput_LostFocus(object sender, RoutedEventArgs evt) {
            InputLostFocus((TextBox)sender, Validator.CheckPredecessorMgNr);
        }

        private new void PhoneNrInput_TextChanged(object sender, RoutedEventArgs evt) {
            base.PhoneNrInput_TextChanged(sender, evt);
            UpdatePhoneNrInputVisibility(IsEditing || IsCreating);
        }
    }
}
